from cpl_core.configuration import ConfigurationABC
from cpl_core.database.context import DatabaseContextABC
from cpl_discord.service import DiscordBotServiceABC

from bot_core.configuration.feature_flags_enum import FeatureFlagsEnum
from bot_core.configuration.feature_flags_settings import FeatureFlagsSettings
from bot_data.abc.achievement_repository_abc import AchievementRepositoryABC
from bot_data.abc.auto_role_repository_abc import AutoRoleRepositoryABC
from bot_data.abc.client_repository_abc import ClientRepositoryABC
from bot_data.abc.game_server_repository_abc import GameServerRepositoryABC
from bot_data.abc.level_repository_abc import LevelRepositoryABC
from bot_data.abc.scheduled_event_repository_abc import ScheduledEventRepositoryABC
from bot_data.abc.server_config_repository_abc import ServerConfigRepositoryABC
from bot_data.abc.short_role_name_repository_abc import ShortRoleNameRepositoryABC
from bot_data.abc.user_joined_server_repository_abc import UserJoinedServerRepositoryABC
from bot_data.abc.user_joined_voice_channel_repository_abc import (
    UserJoinedVoiceChannelRepositoryABC,
)
from bot_data.abc.user_repository_abc import UserRepositoryABC
from bot_data.model.server import Server
from bot_data.model.server_config import ServerConfig
from bot_data.model.server_history import ServerHistory
from bot_graphql.abc.data_query_with_history_abc import DataQueryWithHistoryABC
from bot_graphql.filter.achievement_filter import AchievementFilter
from bot_graphql.filter.auto_role_filter import AutoRoleFilter
from bot_graphql.filter.client_filter import ClientFilter
from bot_graphql.filter.level_filter import LevelFilter
from bot_graphql.filter.scheduled_event_filter import ScheduledEventFilter
from bot_graphql.filter.short_role_name_filter import ShortRoleNameFilter
from bot_graphql.filter.user_filter import UserFilter
from bot_graphql.model.server_statistics import ServerStatistics


class ServerQuery(DataQueryWithHistoryABC):
    def __init__(
        self,
        config: ConfigurationABC,
        bot: DiscordBotServiceABC,
        db: DatabaseContextABC,
        auto_roles: AutoRoleRepositoryABC,
        clients: ClientRepositoryABC,
        levels: LevelRepositoryABC,
        game_servers: GameServerRepositoryABC,
        users: UserRepositoryABC,
        ujs: UserJoinedServerRepositoryABC,
        ujvs: UserJoinedVoiceChannelRepositoryABC,
        achievements: AchievementRepositoryABC,
        short_role_names: ShortRoleNameRepositoryABC,
        scheduled_events: ScheduledEventRepositoryABC,
        server_configs: ServerConfigRepositoryABC,
    ):
        DataQueryWithHistoryABC.__init__(self, "Server", "ServersHistory", ServerHistory, db)

        self._config = config
        self._bot = bot
        self._auto_roles = auto_roles
        self._clients = clients
        self._levels = levels
        self._users = users
        self._ujs = ujs
        self._ujvs = ujvs

        self.set_field("id", self.resolve_id)
        self.set_field("discordId", self.resolve_discord_id)
        self.set_field("name", self.resolve_name)
        self.set_field("iconURL", self.resolve_icon_url)

        self.add_collection(
            "autoRole",
            lambda server, *_: self._auto_roles.get_auto_roles_by_server_id(server.id),
            AutoRoleFilter,
        )
        self.add_collection(
            "client",
            lambda server, *_: self._clients.get_clients_by_server_id(server.id),
            ClientFilter,
        )
        self.add_collection(
            "level",
            lambda server, *_: self._levels.get_levels_by_server_id(server.id).order_by_descending(lambda x: x.min_xp),
            LevelFilter,
        )
        self.set_field(
            "activeUserCount",
            lambda server, *_: self._users.get_users_by_server_id(server.id).where(lambda x: not x.left_server).count(),
        )
        self.add_collection(
            "user",
            lambda server, *_: self._users.get_users_with_activity_by_server_id(server.id),
            UserFilter,
        )
        self.add_collection(
            "gameServer",
            lambda server, *_: game_servers.get_game_servers_by_server_id(server.id),
        )
        self.add_collection(
            "achievement",
            lambda server, *_: achievements.get_achievements_by_server_id(server.id),
            AchievementFilter,
        )
        self.add_collection(
            "shortRoleName",
            lambda server, *_: short_role_names.get_short_role_names_by_server_id(server.id),
            ShortRoleNameFilter,
        )
        self.add_collection(
            "scheduledEvent",
            lambda server, *_: scheduled_events.get_scheduled_events_by_server_id(server.id),
            ScheduledEventFilter,
        )
        self.set_field(
            "config",
            lambda server, *_: server_configs.get_server_config_by_server(server.id),
        )
        self.set_field(
            "hasFeatureFlag",
            lambda server, *_, **kwargs: self._resolve_has_feature_flag(server, *_, **kwargs),
        )
        self.set_field(
            "featureFlags",
            lambda server, *_, **kwargs: self._resolve_feature_flags(server, *_),
        )
        self.set_field("statistic", lambda server, *_, **kwargs: ServerStatistics(server, kwargs))

    @staticmethod
    def resolve_id(server: Server, *_):
        return server.id

    @staticmethod
    def resolve_discord_id(server: Server, *_):
        return server.discord_id

    @staticmethod
    def resolve_name(server: Server, *_):
        return server.name

    @staticmethod
    def resolve_icon_url(server: Server, *_):
        return server.icon_url

    def _resolve_has_feature_flag(self, server: Server, *_, **kwargs):
        settings: ServerConfig = self._config.get_configuration(f"ServerConfig_{server.discord_id}")
        if "flag" not in kwargs:
            return False
        return {
            "key": kwargs["flag"],
            "value": FeatureFlagsSettings.get_flag_from_dict(settings.feature_flags, FeatureFlagsEnum(kwargs["flag"])),
        }

    def _resolve_feature_flags(self, server: Server, *_) -> list[dict]:
        settings: ServerConfig = self._config.get_configuration(f"ServerConfig_{server.discord_id}")
        return [
            {
                "key": flag,
                "value": FeatureFlagsSettings.get_flag_from_dict(settings.feature_flags, FeatureFlagsEnum(flag)),
            }
            for flag in [e.value for e in FeatureFlagsEnum]
        ]
